/* ia32-function.c - ia32 function abstraction layer
 *
 * by scut
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <common.h>
#include <ia32-function.h>


/*** static prototypes
 */
static void
ia32_func_treeplain_fixreloc (ia32_function *func, elf_reloc_list *rel,
	ia32_function **flist, unsigned int flist_count);

static ia32_branch **
ia32_func_br_sp_list_2 (ia32_branch **btl, unsigned int btl_len,
	ia32_function *func, unsigned int *brl_len,
	ia32_branch *source, ia32_branch *dest);

static ia32_branch *
ia32_func_br_sp_list_3 (ia32_branch **brl, unsigned int brl_len,
	unsigned int idx, int *sparr);

static void
ia32_func_br_sp_2 (ia32_branch **brl, unsigned int brl_length,
	unsigned int brl_cur, int *patharray);

static void
ia32_func_breakup_2 (ia32_function *func, ia32_branch *root,
	ia32_branch *this, elf_reloc_list *rel);

static void
ia32_graphviz_func_out_2 (FILE *fp, ia32_function **flist,
	unsigned int flist_count, int *flist_consider);

/*** implementation
 */

ia32_xref *
ia32_xref_new (void)
{
	return (xcalloc (1, sizeof (ia32_xref)));
}


ia32_function *
ia32_func_new (void)
{
	return (xcalloc (1, sizeof (ia32_function)));
}


void
ia32_func_xref_add (ia32_function *func, int totype,
	unsigned int at, unsigned int addend, unsigned int to,
	ia32_function *called, Elf32_Rel *rel)
{
	ia32_xref *	xref;


	xref = ia32_xref_new ();
	if (rel != NULL)
		memcpy (&xref->orig, rel, sizeof (Elf32_Rel));

	xref->from = at;
	xref->addend = addend;
	xref->to = to;
	xref->to_type = totype;
	xref->to_data = called;

	func->func_xref_count += 1;
	func->func_xref = xrealloc (func->func_xref, func->func_xref_count *
		sizeof (ia32_xref *));
	func->func_xref[func->func_xref_count - 1] = xref;

	return;
}


ia32_xref *
ia32_func_xref_findfrom (ia32_function *func, unsigned int vaddr_from)
{
	unsigned int	xrn;


	for (xrn = 0 ; xrn < func->func_xref_count ; ++xrn) {
		if (func->func_xref[xrn]->from == vaddr_from)
			return (func->func_xref[xrn]);
	}

	return (NULL);
}


unsigned int
ia32_func_xref_count (ia32_function **flist, unsigned int flist_count,
	int xref_type)
{
	unsigned int	fln,
			xrn,
			xref_count = 0;


	for (fln = 0 ; fln < flist_count ; ++fln) {
		for (xrn = 0 ; xrn < flist[fln]->func_xref_count ; ++xrn) {
			if (flist[fln]->func_xref[xrn]->to_type == xref_type)
				xref_count += 1;
		}
	}

	return (xref_count);
}


ia32_branch *
ia32_func_find_branch_byrelofs (ia32_function *func, unsigned int relofs)
{
	unsigned int	n;
	unsigned int	brall_count;
	ia32_branch **	brall;
	ia32_branch *	br;


	brall = ia32_br_get_all (func->br_root, &brall_count);

	for (n = 0 ; n < brall_count ; ++n) {
		if ((brall[n]->start - func->start) > relofs)
			continue;
		if ((brall[n]->end - func->start) <= relofs)
			continue;

		br = brall[n];
		free (brall);

		return (br);
	}

	free (brall);

	return (NULL);
}


void
ia32_func_list_walk (ia32_function **flist, unsigned int flist_count,
	void (* walk)(ia32_function *))
{
	unsigned int	n;


	for (n = 0 ; n < flist_count ; ++n)
		walk (flist[n]);

	return;
}


/* br_xref_fromfunc
 *
 * sort in all other relocations happening within function `func' into the
 * corresponding branches for easier/faster access on branch level.
 *
 * return in any case
 */

void
ia32_func_oxref_fromfunc (ia32_function *func)
{
	unsigned int	n,
			dst;
	ia32_branch *	brc;
	ia32_xref **	xrarr;


	/* sort in all other cross references into their corresponding branches
	 */
	for (n = 0 ; n < func->other_xref_count ; ++n) {
		assert (func->other_xref[n] != NULL);
		brc = ia32_func_find_branch_byrelofs (func,
			func->other_xref[n]->from);

		assert (brc != NULL);
		brc->other_xref_count += 1;
		brc->other_xref = xrealloc (brc->other_xref,
			brc->other_xref_count * sizeof (ia32_xref *));
		xrarr = brc->other_xref;
		xrarr[brc->other_xref_count - 1] = func->other_xref[n];
		func->other_xref[n] = NULL;
	}

	/* compress now modified func->other_xref array
	 */
	for (dst = 0, n = 0 ; n < func->other_xref_count ; ++n) {
		if (func->other_xref[n] == NULL)
			continue;

		func->other_xref[dst] = func->other_xref[n];
		dst += 1;
	}

	func->other_xref_count = dst;
	if (dst == 0) {
		free (func->other_xref);
	} else {
		func->other_xref = xrealloc (func->other_xref, dst *
			sizeof (ia32_xref *));
	}

	return;
}


unsigned char *
ia32_func_v2real (ia32_function *func, unsigned int vaddr)
{
	/* FIXME: look into dietlibc issue of having all functions zero byted
	 */
	assert (vaddr >= func->start && vaddr < func->end);

	return (&func->mem[vaddr - func->start]);
}


unsigned int
ia32_func_r2virt (ia32_function *func, unsigned int real)
{
	unsigned char *	rptr = (unsigned char *) real;

	assert (rptr >= func->mem &&
		rptr < (func->mem + (func->end - func->start)));

	return (rptr - func->mem);
}


void
ia32_func_treeplain (ia32_function **flist, unsigned int flist_count,
	elf_reloc_list *rel)
{
	unsigned int	n;
	ia32_function *	func;

	/* collect all branches of the function
	 */
	fprintf (stderr, "\n");

	while ((func = ia32_func_get_unpassed (flist, flist_count)) != NULL) {
		fprintf (stderr, "\t%-32s", func->name);

		func->br_root = ia32_func_breakup (func, rel);
		fprintf (stderr, "%-10u\n", func->func_xref_count);
		func->passed = 1;
	}

	/* fix cross references
	 */
	for (n = 0 ; n < flist_count ; ++n)
		ia32_func_treeplain_fixreloc (flist[n], rel,
			flist, flist_count);

	return;
}


static void
ia32_func_treeplain_fixreloc (ia32_function *func, elf_reloc_list *rel,
	ia32_function **flist, unsigned int flist_count)
{
	unsigned int	xc,
			rn;
	ia32_xref *	xref;
	elf_reloc *	reloc;
	unsigned int	rfunc_offset;	/* relative to function offset */


	for (xc = 0 ; xc < func->func_xref_count ; ++xc) {
		xref = func->func_xref[xc];

		/* in case its refering an external function we cannot provide
		 * the ia32_function pointer, so set it to NULL, humm...
		 */
		if (xref->to_type == IA32_XREF_FUNCEXTERN) {
			fprintf (stderr, "extern function reference from "
				"0x%08x (%s)\n", xref->from, func->name);

			xref->to_data = NULL;

			continue;
		} else if (xref->to_type == IA32_XREF_FUNCTION) {
			xref->to_data = ia32_func_list_find_bystart (flist,
				flist_count, xref->to);
		} else {
			assert (0);
		}

		assert (xref->to_data != NULL);
	}

	/* now add the non-directly-control flow function cross references.
	 * for example code like this:
	 *      push   function
	 *      call   otherfunc
	 * then a reloc will be emitted by the linker that will insert the
	 * address of function at the correct place. since we do overlays,
	 * we have to insert the address in runtime and hence need this info.
	 */
	for (xc = 0 ; xc < rel->reloc_count ; ++xc) {
		reloc = rel->reloc[xc];

		if (reloc->offset_rel < func->start ||
			reloc->offset_rel >= func->end)
			continue;

		/* the crossreferences always refer relative to the function
		 * start
		 */
		rfunc_offset = reloc->offset_rel - func->start;

		/* check whether we already have this xref in the function
		 * references
		 */
		for (rn = 0 ; rn < func->func_xref_count ; ++rn) {
			if ((func->func_xref[rn]->from +
				func->func_xref[rn]->addend) == rfunc_offset)
			{
				break;
			}
		}

		if (rn < func->func_xref_count)
			continue;

		/* now we've found a relocation that is happening within the
		 * function body. we add it, since it is currently missing (it
		 * was not mentioned by any control flow instruction from
		 * within the functions code)
		 */
		ia32_func_oxref_add (func, rfunc_offset, reloc);
	}

}


void
ia32_func_oxref_add (ia32_function *func, unsigned int relofs, elf_reloc *rel)
{
	ia32_xref *	xref = ia32_xref_new ();


	memcpy (&xref->orig, &rel->orig, sizeof (Elf32_Rel));

	/* no addend for other cross references
	 */
	xref->from = relofs;
	xref->addend = 0;
	xref->to_type = IA32_XREF_OTHER;
	xref->to_data = (void *) rel;
	xref->rel_addend = *((unsigned int *) &func->mem[relofs]);

	func->other_xref_count += 1;
	func->other_xref = xrealloc (func->other_xref, func->other_xref_count *
		sizeof (ia32_xref *));
	func->other_xref[func->other_xref_count - 1] = xref;

	return;
}


ia32_function *
ia32_func_get_unpassed (ia32_function **flist, unsigned int flist_count)
{
	unsigned int	n;


	for (n = 0 ; n < flist_count ; ++n)
		if (flist[n]->passed == 0)
			return (flist[n]);

	return (NULL);
}


ia32_branch *
ia32_func_br_end (ia32_function *func)
{
	int		n;
	ia32_branch *	br_end;
	ia32_branch **	brl;


	brl = ia32_br_get_all (func->br_root, (unsigned int *) &n);

	for (n = n - 1 ; n >= 0 ; --n)
		if (brl[n]->endtype == BR_END_RET)
			break;
	assert (n >= 0);
	br_end = brl[n];
	free (brl);

	return (br_end);
}


ia32_branch **
ia32_func_br_mustexec (ia32_function *func, unsigned int *br_count)
{
	unsigned int	n,
			bn;
	ia32_branch *	br_return;

	ia32_branch **	br_all;
	unsigned int	br_all_count;

	ia32_branch **	brl_sp;
	unsigned int	brl_sp_count;

	ia32_branch **	brl_must = NULL;
	unsigned int	brl_must_count = 0;


	/* 1. branches that must be executed on every functions run have to lie
	 * on the shortest path
	 */
	br_return = ia32_func_br_end (func);
	brl_sp = ia32_func_br_sp_list (func, &brl_sp_count,
		func->br_root, br_return);
	assert (brl_sp_count > 0);

	/* the startnode must lie in the list for sure
	 */
	brl_must_count += 1;
	brl_must = xrealloc (brl_must, brl_must_count * sizeof (ia32_branch));
	brl_must[brl_must_count - 1] = func->br_root;

	/* 2. for the nodes that lie on the shortest path - except the start
	 * and end node, there must be no path around them, else they do not
	 * belong to the must-exec nodes
	 */

	br_all = ia32_br_get_all (func->br_root, &br_all_count);

	for (n = 0 ; n < brl_sp_count ; ++n) {
		ia32_branch *	br_temp;
		ia32_branch **	sp_temp;


		/* ignore start and end, they must be executed anytime
		 */
		if (brl_sp[n] == func->br_root || brl_sp[n] == br_return)
			continue;

		/* temporarily remove branch from the "all branches" list,
		 * then check if there is a path now, with the branch removed
		 */
		for (bn = 0 ; bn < br_all_count ; ++bn)
			if (brl_sp[n] == br_all[bn])
				break;
		assert (bn < br_all_count);

		br_temp = br_all[bn];
/*		assert ((br_all_count - bn - 1) > 0);*/

		br_all_count -= 1;
		memmove (&br_all[bn], &br_all[bn + 1], (br_all_count - n) *
			sizeof (ia32_branch *));

		/* check whether there is a path
		 */
		sp_temp = ia32_func_br_sp_list_2 (br_all, br_all_count, func,
			NULL, func->br_root, br_return);

		if (sp_temp != NULL)
			free (sp_temp);

		/* no path => the branch must be travelled in any possible case
		 */
		if (sp_temp == NULL) {
			brl_must_count += 1;
			brl_must = xrealloc (brl_must,
				brl_must_count * sizeof (ia32_branch));
			brl_must[brl_must_count - 1] = br_temp;
		}

		/* fixup the branch array by inserting the removed branch again
		 */
		memmove (&br_all[bn + 1], &br_all[bn], (br_all_count - n) *
			sizeof (ia32_branch *));
		br_all[bn] = br_temp;
		br_all_count += 1;
	}

	free (brl_sp);
	free (br_all);

	/* the endnode also lies always in the list, in case it is not
	 * identical with the start node
	 */
	if (br_return != func->br_root) {
		brl_must_count += 1;
		brl_must = xrealloc (brl_must, brl_must_count *
			sizeof (ia32_branch));
		brl_must[brl_must_count - 1] = br_return;
	}

	*br_count = brl_must_count;

	return (brl_must);
}


ia32_branch **
ia32_func_br_sp_list (ia32_function *func, unsigned int *brl_len,
	ia32_branch *source, ia32_branch *dest)
{
	ia32_branch **	btl;
	unsigned int	btl_len;
	ia32_branch **	br_ret;

	btl = ia32_br_get_all (func->br_root, &btl_len);
	br_ret = ia32_func_br_sp_list_2 (btl, btl_len, func, brl_len,
		source, dest);

	free (btl);

	return (br_ret);
}


static ia32_branch **
ia32_func_br_sp_list_2 (ia32_branch **btl, unsigned int btl_len,
	ia32_function *func, unsigned int *brl_len,
	ia32_branch *source, ia32_branch *dest)
{
	unsigned int	source_idx,
			dest_idx;
	int *		sp_arr;

	ia32_branch **	sp_bra = NULL;
	unsigned int	sp_idx;
	unsigned int	brl_len_dummy;


	if (btl_len == 0)
		return (NULL);

	if (brl_len == NULL)
		brl_len = &brl_len_dummy;

	source_idx = ia32_brl_find (btl, btl_len, source);
	dest_idx = ia32_brl_find (btl, btl_len, dest);
	assert (source_idx != -1 && dest_idx != -1);

	sp_arr = xcalloc (btl_len, sizeof (int));
	ia32_func_br_sp (btl, btl_len, source_idx, sp_arr);

	/* ensure there is a path from source to dest, else bail out
	 */
	if (sp_arr[dest_idx] == -1)
		goto bail;

	sp_bra = xcalloc (btl_len, sizeof (ia32_branch *));
	sp_idx = btl_len - 1;
	sp_bra[sp_idx] = dest;

	while (sp_bra[sp_idx] != source) {
		sp_idx -= 1;

		sp_bra[sp_idx] = ia32_func_br_sp_list_3 (btl, btl_len,
			dest_idx, sp_arr);

		dest_idx = ia32_brl_find (btl, btl_len, sp_bra[sp_idx]);
		assert (dest_idx != -1);
	}

	*brl_len = btl_len - sp_idx;
	memmove (&sp_bra[0], &sp_bra[sp_idx], *brl_len *
		sizeof (ia32_branch *));
	sp_bra = xrealloc (sp_bra, *brl_len * sizeof (ia32_branch *));

bail:
	free (sp_arr);

	return (sp_bra);
}


static ia32_branch *
ia32_func_br_sp_list_3 (ia32_branch **brl, unsigned int brl_len,
	unsigned int idx, int *sparr)
{
	unsigned int	n,
			bn;


	for (n = 0 ; n < brl_len ; ++n) {
		if (n == idx)
			continue;

		if ((sparr[n] + 1) != sparr[idx])
	      		continue;

		/* now we have a possible node with a correct stepcount, check
		 * whether our node is a subnode of it
		 */
		for (bn = 0 ; bn < brl[n]->endbr_count ; ++bn)
			if (brl[n]->endbr[bn] == brl[idx])
				return (brl[n]);
	}

	assert (n < brl_len);

	return (NULL);
}


void
ia32_func_br_sp (ia32_branch **brl, unsigned int brl_length,
	unsigned int brl_source, int *patharray)
{
	unsigned int	n;


	/* use a recursing bellman-ford algorithm with constant edge weight of
	 * one. recursive due to adjascent vertex restriction in original
	 * bellman-ford.
	 *
	 * 1. initialization, -1 denotes infinity
	 */
	for (n = 0 ; n < brl_length ; ++n)
		patharray[n] = -1;
	patharray[brl_source] = 0;

	ia32_func_br_sp_2 (brl, brl_length, brl_source, patharray);
}


static void
ia32_func_br_sp_2 (ia32_branch **brl, unsigned int brl_length,
	unsigned int brl_cur, int *patharray)
{
	unsigned int	n,
			bidx;
	ia32_branch *	br = brl[brl_cur];


	for (n = 0 ; n < br->endbr_count ; ++n) {
		bidx = ia32_brl_find (brl, brl_length, br->endbr[n]);
		if (bidx == -1 || bidx == brl_cur)
			continue;

		/* 2. relax for each subnode, then recurse
		 */
		if (patharray[bidx] == -1 ||
			patharray[bidx] > (patharray[brl_cur] + 1))
		{
			fprintf (stderr, "relax: 0x%08x from %2d to %2d\n",
				brl[bidx]->start, patharray[bidx],
				patharray[brl_cur] + 1);

			patharray[bidx] = patharray[brl_cur] + 1;

			ia32_func_br_sp_2 (brl, brl_length,
				bidx, patharray);
		}
	}
}


ia32_branch *
ia32_func_breakup (ia32_function *func, elf_reloc_list *rel)
{
	ia32_branch *		root;
	ia32_branch *		cur = ia32_br_new ();


	cur->start = func->start;
	root = cur;

	/* work until all branches are passed
	 */
	while ((cur = ia32_br_get_unpassed (root)) != NULL)
		ia32_func_breakup_2 (func, root, cur, rel);


	return (root);
}


static void
ia32_func_breakup_2 (ia32_function *func, ia32_branch *root,
	ia32_branch *this, elf_reloc_list *rel)
{
	ia32_instruction *	inst,	/* helper pointer for inst_s */
				inst_s;	/* instruction structure */
		/* temporary working branches */
	ia32_branch *		work1;	/* temporary work branch */
	ia32_branch *		work2;	/* temporary work branch */
	unsigned int		vaddr;	/* current virtual address */
	unsigned int		destaddr;	/* control flow destination */
	unsigned char *		cur;	/* memory position of vaddr */
	unsigned int		ilen;	/* instruction length */

	/* flags for a single control flow change */
	int			intra;	/* is within current function? */
	int			cond;	/* is condition-based? */
	int			resume;	/* will the call/jmp will resume? */
	unsigned int		swmask;	/* switchmask */
	int			duesplit;
	int			c_volatile;


	vaddr = this->end = this->start;
	this->passed = 1;
	ilen = 0;

	while (vaddr < func->end) {
		/* relocate and decode current instruction
		 */
		cur = &func->mem[vaddr - func->start];
		inst = ia32_decode_instruction (cur, &inst_s);

		if (inst == NULL) {
			fprintf (stderr, "failed to trace branch at 0x%08x, "
				"in function \"%s\"\n",
				vaddr, func->name == NULL ? "__unknown" :
				func->name);
			this->endtype = BR_END_INVALID;

			return;
		}
		ilen = inst->length;

		this->end = vaddr;


		/* we have hit a control instruction, lets get its target, if
		 * possible.
		 */
		if (OD_TEST (inst->opc.used, OP_CONTROL))
			destaddr = ia32_trace_control (inst, cur, vaddr,
				&resume, &c_volatile);

		/* control flow instructions that cannot be predicted but
		 * resume the code flow are treated as if they are no c/f
		 * instructions. so, after throwing them all in one bucked,
		 * just continue processing after the instruction
		 */
		if ((c_volatile && resume) ||
			OD_TEST (inst->opc.used, OP_CONTROL) == 0)
		{
			vaddr += ilen;

			work1 = ia32_br_find (root, vaddr);
			if (work1 != NULL) {
				this->end = vaddr;
				this->endtype = BR_END_PASS;
				this->endbr = xcalloc (1, sizeof (ia32_branch *));
				this->endbr[0] = work1;
				this->endbr_count = 1;

				return;
			}

			continue;
		}

		/* the last instruction length is stored to be able to morph
		 * the last branch instruction (jcc, call, ret, ..) without
		 * having to re-disassemble it
		 */
		this->last_ilen = ilen;

		intra = 0;
		if (destaddr != 0xffffffff)
			intra = ia32_trace_range (func->start, func->end,
				destaddr);

		cond = (OD_TEST (inst->opc.used, OP_COND)) ? 1 : 0;
#define	SW_INTRA	1
#define	SW_INTER	0
#define	SW_COND		2
#define	SW_UNCOND	0
#define	SW_RESUME	4
#define	SW_PASS		0
		swmask = (intra ? SW_INTRA : 0) | (cond ? SW_COND : 0) |
			(resume ? SW_RESUME : 0);

		this->end = vaddr + ilen;

		/* it is a jmp that does not return, so use _END_UNPREDICT
		 */
		if (c_volatile && resume == 0) {
			this->endtype = BR_END_UNPREDICT;

			return;
		}

		switch (swmask) {
		/* inter function references
		 */
		/* branch terminate, return
		 */
		case (SW_INTER | SW_UNCOND | SW_PASS):
			this->endtype = BR_END_RET;

			return;

		/* branch ignore, conditional jump inter-function
		 */
		case (SW_INTER | SW_COND | SW_PASS):
			break;

		/* branch call, external function reference
		 * add the special passing branch type "call".
		 */
		case (SW_INTER | SW_UNCOND | SW_RESUME):
		case (SW_INTER | SW_COND | SW_RESUME):
			ia32_func_xref_add (func, IA32_XREF_FUNCTION,
				vaddr - func->start, 1,
				destaddr, NULL, NULL);

			this->end = vaddr + ilen;
			work1 = ia32_br_find (root, vaddr + ilen);

			this->endtype = BR_END_CALL;
			this->endbr = xcalloc (1, sizeof (ia32_branch *));

			this->endbr[0] = work1;
			if (this->endbr[0] == NULL) {
				this->endbr[0] = ia32_br_new ();
				this->endbr[0]->start = vaddr + ilen;
			}
			this->endbr_count = 1;

			return;

		/* intra function references
		 */
		/* branch if, conditional jump
		 */
		case (SW_INTRA | SW_COND | SW_PASS):
			if (destaddr == 0xffffffff)
				break;

			/* when the jcc points to within its own instruction,
			 * we can be sure there will be a relocation offset
			 * inserted at that place. this is rarely used, the
			 * only place i have seen it yet was the handling of
			 * system calls in glibc, where an inter-function jump
			 * to __syscall_error is used.
			 */
			if (ia32_trace_range (vaddr, vaddr + ilen, destaddr)) {
				unsigned int	odaddr = destaddr;


				destaddr = elf_reloc_list_lookup (rel,
					destaddr);

				fprintf (stderr, "calljump: to 0x%08x\n",
					destaddr);

				assert (destaddr != 0xffffffff);

				ia32_func_xref_add (func, IA32_XREF_FUNCTION,
					vaddr - func->start, odaddr - vaddr,
					destaddr, NULL, NULL);

				work1 = ia32_br_find (root, vaddr + ilen);

				this->endtype = BR_END_IF_INTER;
				this->cond = inst->opc.cond;
				this->endbr_count = 1;
				this->endbr = xcalloc (1, sizeof (ia32_branch *));
				this->endbr[0] = work1;

				if (this->endbr[0] == NULL) {
					this->endbr[0] = ia32_br_new ();
					this->endbr[0]->start = vaddr + ilen;
				}

				break;
			}

			/* XXX: finding on the vaddr address cannot clash,
			 * since we either find no branch or an already passed
			 * branch. finding on the destination address can split
			 * our current branch (`this'). hence, the first thing
			 * we have to ensure is whether we got a branch and if
			 * resulted through a split
			 */
			duesplit = ia32_trace_range (this->start, this->end,
				destaddr);
			work1 = ia32_br_find (root, vaddr + ilen);
			work2 = ia32_br_find (root, destaddr);

			if (duesplit) {
				assert (work2 != NULL);
				this = work2;
			}

			/* XXX: set this after ia32_br_find, else clash
			 */
			this->endtype = BR_END_IF;
			this->cond = inst->opc.cond;
			this->endbr_count = 2;
			this->endbr = xcalloc (2, sizeof (ia32_branch *));
			this->endbr[0] = work1;
			this->endbr[1] = work2;

			if (this->endbr[0] == NULL) {
				this->endbr[0] = ia32_br_new ();
				this->endbr[0]->start = vaddr + ilen;
			}

			if (this->endbr[1] == NULL) {
				this->endbr[1] = ia32_br_new ();
				this->endbr[1]->start = destaddr;
			}

			return;

		/* branch transfer, unconditional jump
		 */
		case (SW_INTRA | SW_UNCOND | SW_PASS):
			if (destaddr == 0xffffffff)
				break;

			duesplit = ia32_trace_range (this->start, this->end,
				destaddr);
			work1 = ia32_br_find (root, destaddr);

			if (duesplit) {
				assert (work1 != NULL);
				this = work1;
			}

			this->endtype = BR_END_TRANSFER;
			this->endbr = xcalloc (1, sizeof (ia32_branch *));

			this->endbr[0] = work1;
			if (this->endbr[0] == NULL) {
				this->endbr[0] = ia32_br_new ();
				this->endbr[0]->start = destaddr;
			}
			this->endbr_count = 1;

			return;

		/* branch ignore or failure, relocation or non-conform code
		 */
		case (SW_INTRA | SW_UNCOND | SW_RESUME):
		case (SW_INTRA | SW_COND | SW_RESUME):
			/* ignore unknown targets and directly-after-calls
			 * targets, those are used for PIC code
			 */
			if (destaddr == 0xffffffff)
				break;
			if (destaddr == vaddr + ilen)
				break;

			fprintf (stderr, "call at 0x%08x\n", vaddr);

			/* to-be-filled-in relocation positions in ELF object
			 * files have target one after the call instruction
			 * so, look it up in the relocation table.
			 * FIXME: use a more generic "destaddr within
			 *        instruction" approach, ia32_trace_range (..)
			 */
			if (destaddr == vaddr + 1) {
				destaddr = elf_reloc_list_lookup (rel,
					vaddr + 1);

				/* still invalid :-/
				 */
				if (destaddr == 0xffffffff) {
					fprintf (stderr, "failed to obtain "
						"relocation happening at "
						"0x%08x\n", vaddr);

					ia32_func_xref_add (func,
						IA32_XREF_FUNCEXTERN,
						vaddr - func->start, 1,
						destaddr, NULL, NULL);
					break;
				}

				fprintf (stderr, "through reloc at 0x%08x to "
					"0x%08x\n", vaddr, destaddr);
				ia32_func_xref_add (func, IA32_XREF_FUNCTION,
					vaddr - func->start, 1, destaddr, NULL,
					NULL);

				work1 = ia32_br_find (root, vaddr + ilen);

				this->endtype = BR_END_CALL;
				this->endbr = xcalloc (1, sizeof (ia32_branch *));

				this->endbr[0] = work1;
				if (this->endbr[0] == NULL) {
					this->endbr[0] = ia32_br_new ();
					this->endbr[0]->start = vaddr + ilen;
				}
				this->endbr_count = 1;

				fprintf (stderr, "XXX2 DEBUG: call at 0x%08x "
					"to 0x%08x\n", vaddr, destaddr);

				return;
			}

			/* one legit destaddr is the start of this function,
			 * in case its direct recursive. check this case
			 */
			if (func->start == destaddr) {
				/* do not specify an ia32_function structure
				 * for now, will fill that in later
				 */
				ia32_func_xref_add (func, IA32_XREF_FUNCTION,
					vaddr - func->start, 1, destaddr, NULL,
					NULL);

				break;
			}

			/* everything else is an error or complex function
			 * unrolling we cannot cope with
			 */
			fprintf (stderr, "FATAL: complex function hierarchy\n");
			fprintf (stderr, "\tfunc-entry: 0x%08x, "
				"vaddr: 0x%08x, destaddr: 0x%08x",
				func->start, vaddr, destaddr);

			this->endtype = BR_END_INVALID;

			return;

		default:
			break;
		}
#undef	SW_INTRA
#undef	SW_INTER
#undef	SW_COND
#undef	SW_UNCOND
#undef	SW_RESUME
#undef	SW_PASS
		vaddr += ilen;

		/* examine next instruction, is there already a branch for it?
		 */
		work1 = ia32_br_find (root, vaddr);
		if (work1 != NULL) {
			this->end = vaddr;
			this->endtype = BR_END_PASS;
			this->endbr = xcalloc (1, sizeof (ia32_branch *));
			this->endbr[0] = work1;
			this->endbr_count = 1;

			return;
		}
	}
}


ia32_function *
ia32_func_list_find_bystart (ia32_function **flist, unsigned int flist_count,
	unsigned int start)
{
	unsigned int	n;


	for (n = 0 ; n < flist_count ; ++n) {
		if (flist[n]->start == start)
			return (flist[n]);
	}

	return (NULL);
}


ia32_function *
ia32_func_list_find_byname (ia32_function **flist, unsigned int flist_count,
	char *fname)
{
	unsigned int	n;


	for (n = 0 ; n < flist_count ; ++n) {
		if (flist[n]->name == NULL)
			continue;

		if (strcmp (fname, flist[n]->name) == 0)
			return (flist[n]);
	}

	return (NULL);
}


int
ia32_func_list_find_index (ia32_function **flist, unsigned int flist_count,
	ia32_function *func)
{
	unsigned int	n;


	for (n = 0 ; n < flist_count ; ++n)
		if (flist[n] == func)
			return (n);

	return (-1);
}


void
ia32_graphviz_func_out (FILE *fp, ia32_function **flist,
	unsigned int flist_count)
{
	ia32_graphviz_func_out_2 (fp, flist, flist_count, NULL);
}


static void
ia32_graphviz_func_out_2 (FILE *fp, ia32_function **flist,
	unsigned int flist_count, int *flist_consider)
{
	char *		dname;
	ia32_function *	func;
	unsigned int	n,
			i;


	fprintf (fp, "digraph functiondependacies {\n\n");
	fprintf (fp, "\tnode [\n");
	fprintf (fp, "\t\tstyle = filled\n");
	fprintf (fp, "\t\tshape = \"record\"\n");
	fprintf (fp, "\t\tfillcolor = \"lightskyblue\"\n");
	fprintf (fp, "\t];\n\n");

	for (n = 0 ; n < flist_count ; ++n) {
		if (flist_consider != NULL && flist_consider[n] == 0)
			continue;

		fprintf (fp, "\t\"%s\" [\n", flist[n]->name);
		fprintf (fp, "\t\tlabel = \"{ <fi> | %s | <fo> }\"\n",
			flist[n]->name);
		fprintf (fp, "\t];\n");
	}

	for (n = 0 ; n < flist_count ; ++n) {
		if (flist_consider != NULL && flist_consider[n] == 0)
			continue;

		for (i = 0 ; i < flist[n]->func_xref_count ; ++i) {
			if (flist[n]->func_xref[i]->to_type !=
				IA32_XREF_FUNCTION)
				continue;

			func = (ia32_function *) flist[n]->func_xref[i]->to_data;
			if (func == NULL)
				continue;

			if (flist_consider != NULL &&
				flist_consider[ia32_func_list_find_index (flist,
					flist_count, func)] == 0)
				continue;

			dname = func->name;
			if (dname != NULL)
				fprintf (fp, "\t\"%s\":fo -> \"%s\":fi;\n",
					flist[n]->name, dname);
		}
	}

	fprintf (fp, "}\n");

	return;
}


void
ia32_graphviz_func_out_calltree (FILE *fp, ia32_function **flist,
	unsigned int flist_count, ia32_function *interest)
{
	int			int_idx;
	int *			int_arr;

	ia32_function *		func;
	unsigned int		n_m,
				n,
				i;


	assert (interest != NULL);
	int_idx = ia32_func_list_find_index (flist, flist_count, interest);
	assert (int_idx != -1);

	int_arr = xcalloc (flist_count, sizeof (int));
	int_arr[int_idx] = 1;

	/* do a little by brute force,
	 * FIXME: use a more elegant way ;)
	 */
#if 0
	for (n_m = 0 ; n_m < (flist_count * flist_count) ; ++n_m) {
#endif
	for (n_m = 0 ; n_m < (32 * flist_count) ; ++n_m) {
		n = n_m % flist_count;

		if (int_arr[n] != 0)
			continue;

		for (i = 0 ; i < flist[n]->func_xref_count ; ++i) {
			unsigned int	sub_idx;

			if (flist[n]->func_xref[i]->to_type !=
				IA32_XREF_FUNCTION)
				continue;

			func = (ia32_function *) flist[n]->func_xref[i]->to_data;
			if (func == NULL)
				continue;

			sub_idx = ia32_func_list_find_index (flist,
				flist_count, func);
			if (int_arr[sub_idx] == 1)
				int_arr[n] = 1;
		}
	}

	ia32_graphviz_func_out_2 (fp, flist, flist_count, int_arr);

	free (int_arr);
}


