/* ia32-function.h - ia32 function abstraction layer, include file
 *
 * by scut
 */

#ifndef	IA32_FUNCTION_H
#define	IA32_FUNCTION_H

#define	IA32_FUNCTION_VERSION	"0.1.1"

#include <elf.h>
#include <ia32-trace.h>


/* ia32_xref
 *
 * cross reference structure used for any "a references b" kind of
 * relationships, with 'a' being the `from', 'b' the `to'.
 *
 * functions that operate on crossreferences are not bound to prefixes.
 */
typedef struct {
	Elf32_Rel	orig;	/* original relocation entry */

	unsigned int	from;	/* address: from where its referenced */
	unsigned int	addend;	/*   real reloc is at from + addend */

	unsigned int	to;	/* address: referenced address (target) */
	unsigned int	rel_addend;	/* relocation addend (A) */

#define	IA32_XREF_INVALID	0
#define	IA32_XREF_FUNCTION	1
#define	IA32_XREF_FUNCEXTERN	2
#define	IA32_XREF_OTHER		3
	int		to_type;	/* type of target reference */
	/* to_data = pointer to structure defining the target. can be one
	 * of the following types:
	 *    ia32_function (IA32_XREF_FUNCTION)
	 *    elf_reloc (IA32_XREF_OTHER)
	 */
	void *		to_data;
} ia32_xref;

/* ia32_function
 *
 * central function definition. every function can be defined through one of
 * this structures (and usually just one should be used).
 *
 * functions that operate on function structures are ia32_func_*
 */
typedef struct ia32_function {
	int		passed;		/* 0 = not analyzed, 1 = done */
	char *		name;		/* name of function, or NULL */
	unsigned char *	mem;		/* address of function, if loaded */

	unsigned int	start;		/* virtual start address */
	unsigned int	end;		/* virtual end address */

	ia32_branch *	br_root;	/* root branch (func entry point) */

	/* function cross references: list of functions this function calls
	 * XXX: notice that control flow instructions that refer to other
	 *      functions have the vaddr of the instruction requiring the
	 *      address as `from', while data references to functions (function
	 *      pointers that is) have the exact address the relocation has to
	 *      be filled in.
	 */
	unsigned int	func_xref_count;/* number of calls */
	ia32_xref **	func_xref;	/* calls to other functions */

	unsigned int	other_xref_count;
	ia32_xref **	other_xref;

	/* relocations do be processed within this function body */
	unsigned int	reloc_count;	/* number of relocations */
#if 0
	struct elf_reloc **	reloc;		/* relocations themselves */
#endif
	void *		reloc;
} ia32_function;

#include <elf_reloc.h>


/* ia32_xref_new
 *
 * create a new cross reference structure.
 *
 * return pointer to new structure
 */

ia32_xref *
ia32_xref_new (void);


/* ia32_func_xref_find
 *
 * find crossreferences from `vaddr_from' within function body `func'.
 *
 * return pointer to xref found on success
 * return NULL on failure
 */

ia32_xref *
ia32_func_xref_findfrom (ia32_function *func, unsigned int vaddr_from);


/* ia32_func_new
 *
 * create a new function structure.
 *
 * return pointer to new structure
 */

ia32_function *
ia32_func_new (void);


/* ia32_func_get_unpassed
 *
 * get the first function which is not passed yet. walk through `flist', which
 * is `flist_count' items long.
 *
 * return pointer to unpassed function on success
 * return NULL on failure
 */

ia32_function *
ia32_func_get_unpassed (ia32_function **flist, unsigned int flist_count);


/* ia32_func_br_end
 *
 * find the end branch for the function `func'.
 *
 * return branch pointer on success
 * bail on failure (every function must have an end branch)
 */

ia32_branch *
ia32_func_br_end (ia32_function *func);


/* ia32_func_v2real
 *
 * convert a virtual address `vaddr' that lies within the function `func' to
 * the real memory address this address is mapped to in our process space.
 *
 * return the real address on success
 * bail on failure
 */

unsigned char *
ia32_func_v2real (ia32_function *func, unsigned int vaddr);


/* ia32_func_r2virt
 *
 * convert the real address `real' from within the mapped function `func's body
 * to a relative to the function start address.
 *
 * return address on success
 * bail on failure
 */

unsigned int
ia32_func_r2virt (ia32_function *func, unsigned int real);


/* ia32_func_treeplain
 *
 * branch-analyse all unpassed functions in `flist', which is `flist_count'
 * items long. make them passed afterwards. to reference function call
 * addresses, optionally use `rel'. when `rel' is non-NULL it is the
 * relocation list for the section the functions are located in.
 *
 * return in any case
 */

void
ia32_func_treeplain (ia32_function **flist, unsigned int flist_count,
	elf_reloc_list *rel);


/* ia32_func_breakup
 *
 * break a function into a branch tree, starting at `func->start'. use `rel'
 * when non-NULL, like in ia32_func_treedown.
 *
 * return branch root node for function on success
 * should not fail
 */

ia32_branch *
ia32_func_breakup (ia32_function *func, elf_reloc_list *rel);


/* ia32_func_br_mustexec
 *
 * locate the branches of the function `func' that must be executed in any
 * possible execution path. `br_count' must be non-NULL.
 *
 * return list of branches, `br_count' items long.
 */

ia32_branch **
ia32_func_br_mustexec (ia32_function *func, unsigned int *br_count);


/* ia32_func_br_sp_list
 *
 * search the function `func' for one of the shortest execution path between
 * `source' and `dest'. `brl_len' can be NULL, then just the list is returned,
 * which cannot be walked, only used as a "is there a path" indicator.
 *
 * return the path list, which is `brl_len' items long on success
 * return NULL on failure
 */

ia32_branch **
ia32_func_br_sp_list (ia32_function *func, unsigned int *brl_len,
	ia32_branch *source, ia32_branch *dest);


/* ia32_func_br_sp
 *
 * compute the shortest path array for the branch list `brl', which is
 * `brl_length' items long. use the `brl_source' item as path source and store
 * the edge-length array into `patharray'. when there is no path to a node,
 * the path array entry will contain (-1).
 *
 * return list of branches in one of the shortest pathes, `br_count' items long
 */

void
ia32_func_br_sp (ia32_branch **brl, unsigned int brl_length,
	unsigned int brl_source, int *patharray);


/* ia32_func_xref_count
 *
 * count the number of cross references of type `xref_type' appearing in the
 * whole function list `flist', which is `flist_count' items long.
 *
 * return number of cross references counted (zero if none)
 */

unsigned int
ia32_func_xref_count (ia32_function **flist, unsigned int flist_count,
	int xref_type);


/* ia32_func_oxref_add
 *
 * add other cross reference to function `func' at relative function body
 * index `relofs'. all further info is pulled from `rel'.
 *
 * return in any case
 */

void
ia32_func_oxref_add (ia32_function *func, unsigned int relofs, elf_reloc *rel);


/* ia32_func_xref_add
 *
 * add a cross reference into the function `func'. the xref was found at `at',
 * goes to `to', and is referencing to the function `called', which can be
 * NULL. do not check whether the xref is already existant in `func'. the type
 * of the reference added is given by `totype'. `addend' + `at' is where the
 * real relocation happens. when `rel' is non-NULL the original Elf32_Rel
 * structure is copied from there into the xref.
 *
 * return in any case
 */

void
ia32_func_xref_add (ia32_function *func, int totype,
	unsigned int at, unsigned int addend, unsigned int to,
	ia32_function *called, Elf32_Rel *rel);


/* ia32_func_find_branch_byrelofs
 *
 * find the branch that covers the relative offset `relofs' from function body
 * start of function `func'.
 *
 * return found branch on success
 * return NULL on failure
 */

ia32_branch *
ia32_func_find_branch_byrelofs (ia32_function *func, unsigned int relofs);


/* br_xref_fromfunc
 *
 * sort in all other relocations happening within function `func' into the
 * corresponding branches for easier/faster access on branch level.
 *
 * return in any case
 */

void
ia32_func_oxref_fromfunc (ia32_function *func);


/* ia32_func_list_find_bystart
 *
 * find a function starting at address `start' from the list `flist', which is
 * `flist_count' items long.
 *
 * return ia32_function * pointer on success
 * return NULL on failure
 */

ia32_function *
ia32_func_list_find_bystart (ia32_function **flist, unsigned int flist_count,
	unsigned int start);


/* ia32_func_list_walk
 *
 * call function `walk' for any function within the function list `flist',
 * which is `flist_count' items long.
 *
 * return in any case
 */

void
ia32_func_list_walk (ia32_function **flist, unsigned int flist_count,
	void (* walk)(ia32_function *));


/* ia32_func_list_find_byname
 *
 * find the functio named `fname' from the function list `flist', which is
 * `flist_count' items long.
 *
 * return pointer to ia32_function structure on success
 * return NULL on failure
 */

ia32_function *
ia32_func_list_find_byname (ia32_function **flist, unsigned int flist_count,
	char *fname);


/* ia32_func_list_find_index
 *
 * convenience function, find the function `func' in the function list `flist',
 * which is `flist_count' items long.
 *
 * return index found on success
 * return -1 on failure
 */

int
ia32_func_list_find_index (ia32_function **flist, unsigned int flist_count,
	ia32_function *func);


/* ia32_graphviz_func_out
 *
 * output the function dependancy relationships described by `flist', which is
 * `flist_count' items long in .dot graphviz format. output to file identified
 * with file pointer `fp', which has to be open already, and will remain open
 * when this function returns.
 *
 * return in any case
 */

void
ia32_graphviz_func_out (FILE *fp, ia32_function **flist,
	unsigned int flist_count);


/* ia32_graphviz_func_out_calltree
 *
 * like ia32_graphviz_func_out, but only consider the calltrees reaching down
 * to function `interest'.
 *
 * return in any case
 */

void
ia32_graphviz_func_out_calltree (FILE *fp, ia32_function **flist,
	unsigned int flist_count, ia32_function *interest);

#endif

